"""Pydantic models for advanced iterator configuration."""

from typing import Literal

from pydantic import BaseModel, Field


class MaskingConfig(BaseModel):
    """Masking configuration.

    Attributes:
        mode (Literal["Table Name", "Label Name"]): Mode of masking to be applied.
            If "Table Name", the identifier refers to a masking table name.
            If "Label Name", the identifier refers to a label image name.
        identifier (str | None): Name of the masking table or label image
            depending on the mode.
    """

    mode: Literal["Table Name", "Label Name"] = "Table Name"
    identifier: str | None = None


class IteratorConfig(BaseModel):
    """Advanced iterator configuration.

    Attributes:
        masking (MaskingConfig | None): If set, the segmentation will be
            performed only within the confines of the specified mask. A mask can be
            specified either by a label image or a Masking ROI table.
        roi_table (str | None): Name of a ROI table. If set, the segmentation
            will be applied to each ROI in the table individually. This option can
            be combined with masking.
    """

    masking: MaskingConfig | None = Field(
        default=None, title="Masking Iterator Configuration"
    )
    roi_table: str | None = Field(default=None, title="Iterate Over ROIs")



"""Standard transforms for fractal tasks."""

from fractal_tasks_utils.transforms._transforms import (
    GaussianBlurConfig,
    HistogramEqualizationConfig,
    MedianFilterConfig,
    SizeFilterConfig,
)

__all__ = [
    "GaussianBlurConfig",
    "MedianFilterConfig",
    "HistogramEqualizationConfig",
    "SizeFilterConfig",
]



"""Pydantic models for advanced iterator configuration."""

import logging
from typing import Annotated, Literal

import numpy as np
from dask import array as da
from ngio.io_pipes._ops_axes import AxesOps
from ngio.io_pipes._ops_slices import SlicingOps
from pydantic import BaseModel, Field
from skimage.exposure import equalize_adapthist
from skimage.filters import gaussian, median
from skimage.morphology import remove_small_objects

logger = logging.getLogger("fractal_tasks_utils.transforms")

# class TransformProtocol(Protocol):
#    """Protocol for a generic transform."""
#
#    def get_as_numpy_transform(
#        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
#    ) -> np.ndarray:
#        """A transformation to be applied after loading a numpy array."""
#        ...
#
#    def get_as_dask_transform(
#        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
#    ) -> da.Array:
#        """A transformation to be applied after loading a dask array."""
#        ...
#
#    def set_as_numpy_transform(
#        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
#    ) -> np.ndarray:
#        """A transformation to be applied before writing a numpy array."""
#        ...
#
#    def set_as_dask_transform(
#        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
#    ) -> da.Array:
#        """A transformation to be applied before writing a dask array."""
#        ...


class GaussianBlurTransform:
    """Gaussian pre-processing configuration."""

    def __init__(self, sigma_xy: float = 2.0, sigma_z: float | None = None):
        self.sigma_xy = sigma_xy
        self.sigma_z = sigma_z

    def apply(self, image: np.ndarray) -> np.ndarray:
        """Apply Gaussian filter to the image.

        Args:
            image (np.ndarray): Input image.

        Returns:
            np.ndarray: Filtered image.
        """
        if image.ndim == 2:
            if self.sigma_z is not None:
                logger.warning(
                    "sigma_z is specified but the input image is 2D. Ignoring sigma_z."
                )
            return gaussian(image, sigma=self.sigma_xy)
        elif image.ndim == 3:
            sigma = (
                self.sigma_z if self.sigma_z is not None else 0,
                self.sigma_xy,
                self.sigma_xy,
            )
            return gaussian(image, sigma=sigma)
        elif image.ndim == 4:
            sigma = (
                0,
                self.sigma_z if self.sigma_z is not None else 0,
                self.sigma_xy,
                self.sigma_xy,
            )
            return gaussian(image, sigma=sigma)
        else:
            raise ValueError("Input to Gaussian filter image must be 2D, 3D, or 4D.")

    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply Gaussian blur transformation to a numpy array."""
        return self.apply(array)

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply Gaussian blur transformation to a dask array."""
        # apply the Gaussian filter to each chunk of the dask array using map_blocks
        raise NotImplementedError(
            "Gaussian blur transformation is not implemented for dask arrays yet."
        )

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Get Gaussian blur transformation applied before writing a numpy array."""
        return array

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Get Gaussian blur transformation applied before writing a dask array."""
        return array


class GaussianBlurConfig(BaseModel):
    """Configuration for Gaussian blur transformation.

    Attributes:
        type (Literal["gaussian"]): Type of transformation.
        sigma_xy (float): Standard deviation for Gaussian kernel in XY plane.
        sigma_z (float | None): Standard deviation for Gaussian kernel in Z axis.
            If not specified, no blurring is applied in Z axis.
    """

    type: Literal["gaussian"] = "gaussian"
    sigma_xy: float = Field(default=2.0, gt=0)
    sigma_z: float | None = None

    def to_transform(self) -> GaussianBlurTransform:
        """Convert the configuration to a GaussianBlurTransform instance."""
        return GaussianBlurTransform(sigma_xy=self.sigma_xy, sigma_z=self.sigma_z)


class MedianFilterTransform:
    """Median filter pre-processing transformation."""

    def __init__(self, size_xy: int = 2, size_z: int | None = None):
        self.size_xy = size_xy
        self.size_z = size_z

    def apply(self, image: np.ndarray) -> np.ndarray:
        """Apply Median filter to the image.

        Args:
            image (np.ndarray): Input image.

        Returns:
            np.ndarray: Filtered image.
        """
        if image.ndim == 2:
            if self.size_z is not None:
                logger.warning(
                    "size_z is specified but the input image is 2D. Ignoring size_z."
                )
            return median(image, footprint=np.ones((self.size_xy, self.size_xy)))
        elif image.ndim == 3:
            size = (
                self.size_z if self.size_z is not None else 1,
                self.size_xy,
                self.size_xy,
            )
            return median(image, footprint=np.ones(size))
        elif image.ndim == 4:
            size = (
                1,
                self.size_z if self.size_z is not None else 1,
                self.size_xy,
                self.size_xy,
            )
            return median(image, footprint=np.ones(size))
        else:
            raise ValueError("Input to median filter image must be 2D, 3D, or 4D.")

    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply Median filter transformation to a numpy array."""
        return self.apply(array)

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply Median filter transformation to a dask array."""
        raise NotImplementedError(
            "Median filter transformation is not implemented for dask arrays yet."
        )

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Get Median filter transformation applied before writing a numpy array."""
        return array

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Get Median filter transformation applied before writing a dask array."""
        return array


class MedianFilterConfig(BaseModel):
    """Configuration for Median filter transformation.

    Attributes:
        type (Literal["median"]): Type of transformation.
        size_xy (int): Size in pixels of the median filter in XY plane.
        size_z (int | None): Size in pixels of the median filter in Z axis.
            If not specified, no filtering is applied in Z axis.
    """

    type: Literal["median"] = "median"
    size_xy: int = Field(default=2, gt=0)
    size_z: int | None = None

    def to_transform(self) -> MedianFilterTransform:
        """Convert the configuration to a MedianFilterTransform instance."""
        return MedianFilterTransform(size_xy=self.size_xy, size_z=self.size_z)


class HistogramEqualizationTransform:
    """Histogram equalization pre-processing transformation."""

    def __init__(
        self,
        kernel_size_xy: int | None = None,
        kernel_size_z: int | None = None,
        clip_limit: float = 0.01,
        nbins: int = 256,
    ):
        self.kernel_size_xy = kernel_size_xy
        self.kernel_size_z = kernel_size_z
        self.clip_limit = clip_limit
        self.nbins = nbins

    def _build_kernel_size(self, image: np.ndarray) -> np.ndarray | None:
        """Build kernel size tuple based on image dimensions."""
        if image.ndim == 2:
            if self.kernel_size_z is not None:
                logger.warning(
                    "kernel_size_z is specified but the input image is 2D. "
                    "Ignoring kernel_size_z."
                )
            kernel_size = (self.kernel_size_xy, self.kernel_size_xy)

        elif image.ndim == 3:
            kernel_size = (
                self.kernel_size_z,
                self.kernel_size_xy,
                self.kernel_size_xy,
            )

        elif image.ndim == 4:
            kernel_size = (
                1,
                self.kernel_size_z,
                self.kernel_size_xy,
                self.kernel_size_xy,
            )
        else:
            raise ValueError("Input to histogram equalization must be 2D, 3D, or 4D.")

        # Return None if any kernel size component is None (use scikit-image defaults)
        # Return np.array only if all values are specified
        return (
            np.array(kernel_size) if all(k is not None for k in kernel_size) else None
        )

    def apply(self, image: np.ndarray) -> np.ndarray:
        """Apply histogram equalization to the image.

        Args:
            image (np.ndarray): Input image.

        Returns:
            np.ndarray: Histogram equalized image.
        """
        kernel_size = self._build_kernel_size(image)

        return equalize_adapthist(
            image,
            kernel_size=kernel_size,
            clip_limit=self.clip_limit,
            nbins=self.nbins,
        )

    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply histogram equalization transformation to a numpy array."""
        return self.apply(array)

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply histogram equalization transformation to a dask array."""
        raise NotImplementedError(
            "Histogram equalization is not implemented for dask arrays yet."
        )

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Get histogram equalization applied before writing a numpy array."""
        return array

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Get histogram equalization applied before writing a dask array."""
        return array


class HistogramEqualizationConfig(BaseModel):
    """Configuration for Contrast Limited Adaptive Histogram Equalization (CLAHE).

    Attributes:
        type (Literal["histogram"]): Type of transformation.
        kernel_size_xy (int | None): Shape of kernel in XY plane.
            By default, kernel_size is 1/8 of image height by 1/8 of its width.
        kernel_size_z (int | None): Shape of kernel in Z axis.
            By default, kernel_size is 1/8 of image height by 1/8 of its width.
        clip_limit (float): Clipping limit, normalized between 0 and 1
            (higher values give more contrast).
        nbins (int): Number of gray bins for histogram ("data range").
    """

    type: Literal["histogram"] = "histogram"
    kernel_size_xy: int | None = None
    kernel_size_z: int | None = None
    clip_limit: float = Field(default=0.01, ge=0, le=1)
    nbins: int = 256

    def to_transform(self) -> HistogramEqualizationTransform:
        """Convert the configuration to a HistogramEqualizationTransform instance."""
        return HistogramEqualizationTransform(
            kernel_size_xy=self.kernel_size_xy,
            kernel_size_z=self.kernel_size_z,
            clip_limit=self.clip_limit,
            nbins=self.nbins,
        )


class SizeFilterTransform:
    """Size filter post-processing transformation."""

    def __init__(self, min_size: int = 0):
        self.min_size = min_size

    def apply(self, labels: np.ndarray) -> np.ndarray:
        """Apply size filtering to the labeled image.

        Args:
            labels (np.ndarray): Labeled image.

        Returns:
            np.ndarray: Size-filtered labeled image.
        """
        return remove_small_objects(labels, max_size=self.min_size)

    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply size filter transformation to a numpy array."""
        return self.apply(array)

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply size filter transformation to a dask array."""
        raise NotImplementedError(
            "Size filter transformation is not implemented for dask arrays yet."
        )

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Get size filter transformation applied before writing a numpy array."""
        return self.apply(array)

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Get size filter transformation applied before writing a dask array."""
        raise NotImplementedError(
            "Size filter transformation is not implemented for dask arrays yet."
        )


class SizeFilterConfig(BaseModel):
    """Configuration for size filter post-processing.

    Attributes:
        type (Literal["size_filter"]): Type of transformation.
        min_size (int): Minimum size in pixels for objects to keep.
    """

    type: Literal["size_filter"] = "size_filter"
    min_size: int = Field(ge=0)

    def to_transform(self) -> SizeFilterTransform:
        """Convert the configuration to a SizeFilterTransform instance."""
        return SizeFilterTransform(min_size=self.min_size)


PreProcess = Annotated[
    GaussianBlurConfig | MedianFilterConfig | HistogramEqualizationConfig,
    Field(discriminator="type"),
]

PostProcess = Annotated[
    SizeFilterConfig,
    Field(discriminator="type"),
]



"""Core computation logic for segmentation tasks."""

import logging
import time
from collections.abc import Callable

import numpy as np
from ngio import ChannelSelectionModel, OmeZarrContainer, open_ome_zarr_container
from ngio.experimental.iterators import MaskedSegmentationIterator, SegmentationIterator
from ngio.images._masked_image import MaskedImage

from fractal_tasks_utils.segmentation._models import (
    IteratorConfig,
    MaskingConfig,
)
from fractal_tasks_utils.segmentation._transforms import (
    SegmentationTransformConfig,
)


def _load_masked_image(
    ome_zarr: OmeZarrContainer,
    masking_configuration: MaskingConfig,
    logger: logging.Logger,
    level_path: str | None = None,
) -> MaskedImage:
    """Load a masked image from an OME-Zarr based on the masking configuration.

    Args:
        ome_zarr: The OME-Zarr container.
        masking_configuration (MaskingConfig): Configuration for masking.
        level_path (str | None): Optional path to a specific resolution level.

    """
    if masking_configuration.mode == "Table Name":
        masking_table_name = masking_configuration.identifier
        masking_label_name = None
    else:
        masking_label_name = masking_configuration.identifier
        masking_table_name = None
    logger.info(f"Using masking with {masking_table_name=}, {masking_label_name=}")

    # Base Iterator with masking
    masked_image = ome_zarr.get_masked_image(
        masking_label_name=masking_label_name,
        masking_table_name=masking_table_name,
        path=level_path,
    )
    return masked_image


def setup_segmentation_iterator(
    zarr_url: str,
    # Segmentation parameters
    channels: list[ChannelSelectionModel],
    output_label_name: str = "segmentation",
    level_path: str | None = None,
    # Iteration parameters
    iterator_configuration: IteratorConfig | None = None,
    segmentation_transform_config: SegmentationTransformConfig | None = None,
    # Other parameters
    overwrite: bool = True,
) -> SegmentationIterator | MaskedSegmentationIterator:
    """Set up the segmentation iterator based on the provided configuration.

    Args:
        zarr_url (str): URL to the OME-Zarr container
        channels (CellposeChannels): Channels to use for segmentation.
            It must contain between 1 and 3 channel identifiers.
        output_label_name (str): Name of the resulting label image.
        level_path (str | None): If the OME-Zarr has multiple resolution levels,
            the level to use can be specified here. If not provided, the highest
            resolution level will be used.
        iterator_configuration (IteratorConfiguration | None): Configuration
            for the segmentation iterator. This can be used to specify masking
            and/or a ROI table.
        segmentation_transform_config (SegmentationTransformConfig | None):
            Configuration for pre- and post-processing transformations. If not
            provided, no additional transformations will be applied.
        custom_model (str | None): Path to a custom Cellpose model. If not
            set, the default "cpsam" model will be used.
        overwrite (bool): Whether to overwrite an existing label image.
            Defaults to True.
    """
    logger = logging.getLogger("fractal_tasks_utils.setup_iterator")
    # Use the first of input_paths
    logger.info(f"{zarr_url=}")

    # Open the OME-Zarr container
    ome_zarr = open_ome_zarr_container(zarr_url)
    logger.info(f"{ome_zarr=}")
    # Validate that the specified channels are present in the image
    # if _skip_segmentation(channels=channels, ome_zarr=ome_zarr):
    #    return None
    logger.info(f"Formatted label name: {output_label_name=}")

    # Derive the label and an get it at the specified level path
    ome_zarr.derive_label(name=output_label_name, overwrite=overwrite)
    label = ome_zarr.get_label(name=output_label_name, path=level_path)
    logger.info(f"Derived label image: {label=}")

    # Set up the appropriate iterator based on the configuration
    if iterator_configuration is None:
        iterator_configuration = IteratorConfig()

    # Determine if we are doing 3D segmentation or 2D
    axes_order = "czyx" if ome_zarr.is_3d else "cyx"
    logger.info(f"Segmenting using {axes_order=}")

    if segmentation_transform_config is None:
        segmentation_transform_config = SegmentationTransformConfig()

    if iterator_configuration.masking is None:
        # Create a basic SegmentationIterator without masking
        image = ome_zarr.get_image(path=level_path)
        logger.info(f"{image=}")
        iterator = SegmentationIterator(
            input_image=image,
            output_label=label,
            channel_selection=channels,
            axes_order=axes_order,
            input_transforms=segmentation_transform_config.to_pre_transforms(),
            output_transforms=segmentation_transform_config.to_post_transforms(),
        )
    else:
        # Since masking is requested, we need to determine load a masking image
        masked_image = _load_masked_image(
            ome_zarr=ome_zarr,
            masking_configuration=iterator_configuration.masking,
            level_path=level_path,
            logger=logger,
        )
        logger.info(f"{masked_image=}")
        # A masked iterator is created instead of a basic segmentation iterator
        # This will do two major things:
        # 1) It will iterate only over the regions of interest defined by the
        #   masking table or label image
        # 2) It will only write the segmentation results within the masked regions
        iterator = MaskedSegmentationIterator(
            input_image=masked_image,
            output_label=label,
            channel_selection=channels,
            axes_order=axes_order,
            input_transforms=segmentation_transform_config.to_pre_transforms(),
            output_transforms=segmentation_transform_config.to_post_transforms(),
        )
    # Make sure that if we have a time axis, we iterate over it
    # Strict=False means that if there no z axis or z is size 1, it will still work
    # If your segmentation needs requires a volume, use strict=True
    iterator = iterator.by_zyx(strict=False)
    logger.info(f"Iterator created: {iterator=}")

    if iterator_configuration.roi_table is not None:
        # If a ROI table is provided, we load it and use it to further restrict
        # the iteration to the ROIs defined in the table
        # Be aware that this is not an alternative to masking
        # but only an additional restriction
        table = ome_zarr.get_generic_roi_table(name=iterator_configuration.roi_table)
        logger.info(f"ROI table retrieved: {table=}")
        iterator = iterator.product(table)
        logger.info(f"Iterator updated with ROI table: {iterator=}")
    return iterator


def compute_segmentation(
    *,
    segmentation_func: Callable[[np.ndarray], np.ndarray],
    iterator: SegmentationIterator | MaskedSegmentationIterator,
) -> None:
    """Core computation loop for applying the segmentation function.

    This function iterates over the image over the specifed patterns in
    the iterator, applies the segmentation function to each chunk of the image,
    and writes the resulting label images back to the OME-Zarr.

    Args:
        segmentation_func: The segmentation function to apply to each image chunk.
            This function should take an image chunk as input and return a label
            image as output.
        iterator: An iterator that yields image chunks and corresponding writers.
    """
    logger = logging.getLogger("fractal_tasks_utils.compute_segmentation")

    # Keep track of the maximum label to ensure unique across iterations
    max_label = 0
    #
    # Core processing loop
    #
    logger.info("Starting processing...")
    run_times: list[float] = []
    num_rois = len(iterator.rois)
    logging_step = max(1, num_rois // 10)
    for it, (input_img, writer) in enumerate(iterator.iter_as_numpy()):
        start_time = time.time()
        label_img = segmentation_func(input_img)
        # Ensure unique labels across different chunks
        label_img = np.where(label_img == 0, 0, label_img + max_label)
        max_label = max(max_label, label_img.max())
        writer(label_img)
        iteration_time = time.time() - start_time
        run_times.append(iteration_time)

        # Only log the progress every logging_step iterations
        if it % logging_step == 0 or it == num_rois - 1:
            avg_time = sum(run_times) / len(run_times)
            logger.info(
                f"Processed ROI {it + 1}/{num_rois} "
                f"(avg time per ROI: {avg_time:.2f} s)"
            )
    return None



"""Measurement utilities for fractal tasks."""

from fractal_tasks_utils.measurement._compute import (
    compute_measurement,
    join_tables,
    setup_measurement_iterator,
)

__all__ = [
    "compute_measurement",
    "join_tables",
    "setup_measurement_iterator",
]



"""This package contains general utilities to build fractal tasks."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("fractal-tasks-utils")
except PackageNotFoundError:
    __version__ = "unknown"



import numpy as np
import pytest
from pydantic import ValidationError

from fractal_tasks_utils.transforms._transforms import (
    GaussianBlurConfig,
    GaussianBlurTransform,
    HistogramEqualizationConfig,
    HistogramEqualizationTransform,
    MedianFilterConfig,
    MedianFilterTransform,
    SizeFilterConfig,
    SizeFilterTransform,
)

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def image_2d():
    rng = np.random.default_rng(42)
    return rng.random((32, 32), dtype=np.float32)


@pytest.fixture
def image_3d():
    rng = np.random.default_rng(42)
    return rng.random((4, 32, 32), dtype=np.float32)


@pytest.fixture
def image_4d():
    rng = np.random.default_rng(42)
    return rng.random((2, 4, 32, 32), dtype=np.float32)


@pytest.fixture
def label_2d():
    """10x10 label array: label 1 = 25px region, label 2 = 3px region."""
    arr = np.zeros((10, 10), dtype=np.int32)
    arr[1:6, 1:6] = 1  # 5x5 = 25 pixels
    arr[8, 8] = 2
    arr[8, 9] = 2
    arr[9, 8] = 2  # 3 pixels
    return arr


# ---------------------------------------------------------------------------
# GaussianBlurTransform
# ---------------------------------------------------------------------------


def test_gaussian_blur_2d(image_2d):
    result = GaussianBlurTransform(sigma_xy=1.0).apply(image_2d)
    assert result.shape == image_2d.shape
    assert not np.array_equal(result, image_2d)


def test_gaussian_blur_3d(image_3d):
    result = GaussianBlurTransform(sigma_xy=1.0, sigma_z=1.0).apply(image_3d)
    assert result.shape == image_3d.shape


def test_gaussian_blur_invalid_ndim():
    arr = np.zeros((2, 2, 4, 32, 32))
    with pytest.raises(ValueError):
        GaussianBlurTransform().apply(arr)


# ---------------------------------------------------------------------------
# GaussianBlurConfig
# ---------------------------------------------------------------------------


def test_gaussian_blur_config_to_transform():
    t = GaussianBlurConfig(sigma_xy=3.0, sigma_z=1.5).to_transform()
    assert isinstance(t, GaussianBlurTransform)
    assert t.sigma_xy == 3.0
    assert t.sigma_z == 1.5


def test_gaussian_blur_config_invalid_sigma():
    with pytest.raises(ValidationError):
        GaussianBlurConfig(sigma_xy=0)
    with pytest.raises(ValidationError):
        GaussianBlurConfig(sigma_xy=-1.0)


# ---------------------------------------------------------------------------
# MedianFilterTransform
# ---------------------------------------------------------------------------


def test_median_filter_2d(image_2d):
    result = MedianFilterTransform(size_xy=3).apply(image_2d)
    assert result.shape == image_2d.shape


def test_median_filter_invalid_ndim():
    arr = np.zeros((2, 2, 4, 32, 32))
    with pytest.raises(ValueError):
        MedianFilterTransform().apply(arr)


# ---------------------------------------------------------------------------
# MedianFilterConfig
# ---------------------------------------------------------------------------


def test_median_filter_config_to_transform():
    t = MedianFilterConfig(size_xy=5, size_z=2).to_transform()
    assert isinstance(t, MedianFilterTransform)
    assert t.size_xy == 5
    assert t.size_z == 2


def test_median_filter_config_invalid_size():
    with pytest.raises(ValidationError):
        MedianFilterConfig(size_xy=0)


# ---------------------------------------------------------------------------
# HistogramEqualizationTransform
# ---------------------------------------------------------------------------


def test_histogram_equalization_2d(image_2d):
    result = HistogramEqualizationTransform().apply(image_2d)
    assert result.shape == image_2d.shape
    assert result.min() >= 0.0
    assert result.max() <= 1.0


def test_histogram_equalization_3d(image_3d):
    result = HistogramEqualizationTransform().apply(image_3d)
    assert result.shape == image_3d.shape


# ---------------------------------------------------------------------------
# HistogramEqualizationConfig
# ---------------------------------------------------------------------------


def test_histogram_equalization_config_to_transform():
    t = HistogramEqualizationConfig(clip_limit=0.05, nbins=128).to_transform()
    assert isinstance(t, HistogramEqualizationTransform)
    assert t.clip_limit == 0.05
    assert t.nbins == 128


def test_histogram_equalization_config_clip_limit_bounds():
    with pytest.raises(ValidationError):
        HistogramEqualizationConfig(clip_limit=-0.1)
    with pytest.raises(ValidationError):
        HistogramEqualizationConfig(clip_limit=1.1)
    # Boundary values must be accepted
    HistogramEqualizationConfig(clip_limit=0.0)
    HistogramEqualizationConfig(clip_limit=1.0)


# ---------------------------------------------------------------------------
# SizeFilterTransform
# ---------------------------------------------------------------------------


def test_size_filter_removes_small_objects(label_2d):
    result = SizeFilterTransform(min_size=5).apply(label_2d)
    assert 1 in result  # 25px object survives
    assert 2 not in result  # 3px object is removed


def test_size_filter_min_size_zero_keeps_all(label_2d):
    result = SizeFilterTransform(min_size=0).apply(label_2d)
    assert 1 in result
    assert 2 in result


# ---------------------------------------------------------------------------
# SizeFilterConfig
# ---------------------------------------------------------------------------


def test_size_filter_config_to_transform():
    t = SizeFilterConfig(min_size=10).to_transform()
    assert isinstance(t, SizeFilterTransform)
    assert t.min_size == 10


def test_size_filter_config_invalid_min_size():
    with pytest.raises(ValidationError):
        SizeFilterConfig(min_size=-1)



import numpy as np
import pytest
from ngio import ChannelSelectionModel, create_empty_ome_zarr, open_ome_zarr_container
from ngio.experimental.iterators import MaskedSegmentationIterator, SegmentationIterator
from ngio.images._masked_image import MaskedImage

from fractal_tasks_utils.segmentation._compute import (
    _load_masked_image,
    compute_segmentation,
    setup_segmentation_iterator,
)
from fractal_tasks_utils.segmentation._models import (
    IteratorConfig,
    MaskingConfig,
)
from fractal_tasks_utils.segmentation._transforms import SegmentationTransformConfig
from fractal_tasks_utils.transforms._transforms import (
    GaussianBlurConfig,
    GaussianBlurTransform,
    MedianFilterConfig,
    SizeFilterConfig,
    SizeFilterTransform,
)

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------

_CHANNELS = [ChannelSelectionModel(mode="label", identifier="DAPI")]


def _make_2d_zarr(tmp_path, name="test.zarr"):
    zarr_path = str(tmp_path / name)
    ome_zarr = create_empty_ome_zarr(
        store=zarr_path,
        shape=(2, 64, 64),
        pixelsize=0.5,
        channels_meta=["DAPI", "GFP"],
        axes_names=["c", "y", "x"],
    )
    return zarr_path, ome_zarr


@pytest.fixture
def ome_zarr_2d(tmp_path):
    """Minimal 2-channel 2D OME-Zarr."""
    zarr_path, _ = _make_2d_zarr(tmp_path)
    return zarr_path


@pytest.fixture
def ome_zarr_3d(tmp_path):
    """Minimal 2-channel 3D OME-Zarr."""
    zarr_path = str(tmp_path / "test3d.zarr")
    create_empty_ome_zarr(
        store=zarr_path,
        shape=(2, 4, 32, 32),
        pixelsize=0.5,
        z_spacing=1.0,
        channels_meta=["DAPI", "GFP"],
        axes_names=["c", "z", "y", "x"],
    )
    return zarr_path


@pytest.fixture
def ome_zarr_with_masking_label(tmp_path):
    """2D OME-Zarr with a masking label ('organoids') containing 2 non-zero regions."""
    zarr_path, ome_zarr = _make_2d_zarr(tmp_path)
    ome_zarr.derive_label(name="organoids", overwrite=True)
    lbl = ome_zarr.get_label("organoids")
    data = lbl.get_array()
    data[5:25, 5:25] = 1
    data[35:55, 35:55] = 2
    lbl.set_array(data)
    lbl.consolidate()
    return zarr_path


@pytest.fixture
def ome_zarr_with_masking_table(tmp_path):
    """2D OME-Zarr with a masking ROI table built from the 'organoids' label."""
    zarr_path, ome_zarr = _make_2d_zarr(tmp_path)
    ome_zarr.derive_label(name="organoids", overwrite=True)
    lbl = ome_zarr.get_label("organoids")
    data = lbl.get_array()
    data[5:25, 5:25] = 1
    data[35:55, 35:55] = 2
    lbl.set_array(data)
    lbl.consolidate()
    masking_table = ome_zarr.build_masking_roi_table("organoids")
    ome_zarr.add_table(name="masking_table", table=masking_table)
    return zarr_path


@pytest.fixture
def ome_zarr_with_roi_table(tmp_path):
    """2D OME-Zarr with an image-level iteration ROI table."""
    zarr_path, ome_zarr = _make_2d_zarr(tmp_path)
    roi_table = ome_zarr.build_image_roi_table()
    ome_zarr.add_table(name="roi_table", table=roi_table)
    return zarr_path


# ---------------------------------------------------------------------------
# MaskingConfig
# ---------------------------------------------------------------------------


def test_masking_configuration_defaults():
    cfg = MaskingConfig()
    assert cfg.mode == "Table Name"
    assert cfg.identifier is None


def test_masking_configuration_label_name_mode():
    cfg = MaskingConfig(mode="Label Name", identifier="nuclei")
    assert cfg.mode == "Label Name"
    assert cfg.identifier == "nuclei"


# ---------------------------------------------------------------------------
# IteratorConfig
# ---------------------------------------------------------------------------


def test_iterator_configuration_defaults():
    cfg = IteratorConfig()
    assert cfg.masking is None
    assert cfg.roi_table is None


def test_iterator_configuration_with_masking():
    cfg = IteratorConfig(masking=MaskingConfig(mode="Table Name"))
    assert isinstance(cfg.masking, MaskingConfig)
    assert cfg.masking.mode == "Table Name"


def test_iterator_configuration_with_roi_table():
    cfg = IteratorConfig(roi_table="FOV_ROI_table")
    assert cfg.roi_table == "FOV_ROI_table"
    assert cfg.masking is None


# ---------------------------------------------------------------------------
# SegmentationTransformConfig
# ---------------------------------------------------------------------------


def test_segmentation_transform_config_empty_defaults():
    cfg = SegmentationTransformConfig()
    assert cfg.pre_process == []
    assert cfg.post_process == []
    assert cfg.to_pre_transforms() == []
    assert cfg.to_post_transforms() == []


def test_segmentation_transform_config_pre_process_round_trip():
    cfg = SegmentationTransformConfig(pre_process=[GaussianBlurConfig(sigma_xy=2.0)])
    result = cfg.to_pre_transforms()
    assert len(result) == 1
    assert isinstance(result[0], GaussianBlurTransform)
    assert result[0].sigma_xy == 2.0


def test_segmentation_transform_config_post_process_round_trip():
    cfg = SegmentationTransformConfig(post_process=[SizeFilterConfig(min_size=50)])
    result = cfg.to_post_transforms()
    assert len(result) == 1
    assert isinstance(result[0], SizeFilterTransform)
    assert result[0].min_size == 50


def test_segmentation_transform_config_mixed_pre_process():
    cfg = SegmentationTransformConfig(
        pre_process=[
            GaussianBlurConfig(sigma_xy=1.0),
            MedianFilterConfig(size_xy=3),
        ]
    )
    result = cfg.to_pre_transforms()
    assert len(result) == 2


# ---------------------------------------------------------------------------
# _load_masked_image
# ---------------------------------------------------------------------------


def test__load_masked_image_by_table_name(ome_zarr_with_masking_table, caplog):
    import logging

    ome_zarr = open_ome_zarr_container(ome_zarr_with_masking_table)
    cfg = MaskingConfig(mode="Table Name", identifier="masking_table")
    logger = logging.getLogger("test")
    masked = _load_masked_image(ome_zarr, cfg, logger)
    assert isinstance(masked, MaskedImage)


def test__load_masked_image_by_label_name(ome_zarr_with_masking_label):
    import logging

    ome_zarr = open_ome_zarr_container(ome_zarr_with_masking_label)
    cfg = MaskingConfig(mode="Label Name", identifier="organoids")
    logger = logging.getLogger("test")
    masked = _load_masked_image(ome_zarr, cfg, logger)
    assert isinstance(masked, MaskedImage)


# ---------------------------------------------------------------------------
# setup_segmentation_iterator
# ---------------------------------------------------------------------------


def test_setup_segmentation_iterator_2d(ome_zarr_2d):
    iterator = setup_segmentation_iterator(ome_zarr_2d, channels=_CHANNELS)
    assert isinstance(iterator, SegmentationIterator)
    assert len(iterator.rois) > 0


def test_setup_segmentation_iterator_3d(ome_zarr_3d):
    iterator = setup_segmentation_iterator(ome_zarr_3d, channels=_CHANNELS)
    assert isinstance(iterator, SegmentationIterator)
    assert len(iterator.rois) > 0


def test_setup_segmentation_iterator_masked_table(ome_zarr_with_masking_table):
    mc = MaskingConfig(mode="Table Name", identifier="masking_table")
    ic = IteratorConfig(masking=mc)
    iterator = setup_segmentation_iterator(
        ome_zarr_with_masking_table, channels=_CHANNELS, iterator_configuration=ic
    )
    assert isinstance(iterator, MaskedSegmentationIterator)
    assert len(iterator.rois) > 0


def test_setup_segmentation_iterator_masked_label(ome_zarr_with_masking_label):
    mc = MaskingConfig(mode="Label Name", identifier="organoids")
    ic = IteratorConfig(masking=mc)
    iterator = setup_segmentation_iterator(
        ome_zarr_with_masking_label, channels=_CHANNELS, iterator_configuration=ic
    )
    assert isinstance(iterator, MaskedSegmentationIterator)
    assert len(iterator.rois) > 0


def test_setup_segmentation_iterator_with_roi_table(ome_zarr_with_roi_table):
    ic = IteratorConfig(roi_table="roi_table")
    iterator = setup_segmentation_iterator(
        ome_zarr_with_roi_table, channels=_CHANNELS, iterator_configuration=ic
    )
    assert isinstance(iterator, SegmentationIterator)
    assert len(iterator.rois) > 0


def test_setup_segmentation_iterator_default_configs(ome_zarr_2d):
    """Passing None for optional configs should not raise."""
    iterator = setup_segmentation_iterator(
        ome_zarr_2d,
        channels=_CHANNELS,
        iterator_configuration=None,
        segmentation_transform_config=None,
    )
    assert isinstance(iterator, SegmentationIterator)


def test_setup_segmentation_iterator_custom_label_name(ome_zarr_2d):
    setup_segmentation_iterator(
        ome_zarr_2d, channels=_CHANNELS, output_label_name="my_label"
    )
    ome_zarr = open_ome_zarr_container(ome_zarr_2d)
    assert "my_label" in ome_zarr.list_labels()


# ---------------------------------------------------------------------------
# compute_segmentation
# ---------------------------------------------------------------------------


def test_compute_segmentation_basic(ome_zarr_2d):
    """Segmentation function is called and results are written without error."""
    iterator = setup_segmentation_iterator(ome_zarr_2d, channels=_CHANNELS)
    call_count = [0]

    def seg_func(img):
        call_count[0] += 1
        return np.zeros_like(img)

    compute_segmentation(segmentation_func=seg_func, iterator=iterator)
    assert call_count[0] == len(iterator.rois)


def test_compute_segmentation_label_uniqueness(ome_zarr_with_masking_table):
    """Labels from different ROIs must be offset so they don't collide."""
    mc = MaskingConfig(mode="Table Name", identifier="masking_table")
    ic = IteratorConfig(masking=mc)
    iterator = setup_segmentation_iterator(
        ome_zarr_with_masking_table, channels=_CHANNELS, iterator_configuration=ic
    )
    # Each ROI returns labels 1 and 2; with 2 ROIs the second should be offset to 3, 4
    call_count = [0]

    def seg_func(img):
        call_count[0] += 1
        result = np.zeros_like(img)
        result[..., 2:5, 2:5] = 1
        result[..., 6:9, 6:9] = 2
        return result

    compute_segmentation(segmentation_func=seg_func, iterator=iterator)
    assert call_count[0] == 2

    ome_zarr = open_ome_zarr_container(ome_zarr_with_masking_table)
    written = ome_zarr.get_label("segmentation").get_array()
    non_zero = sorted(np.unique(written[written > 0]))
    # Labels should be disjoint across the two ROIs (1, 2 from first; 3, 4 from second)
    assert len(non_zero) == len(set(non_zero))


def test_compute_segmentation_zero_background_preserved(ome_zarr_2d):
    """Zero-valued (background) pixels must remain 0 after the label offset."""
    iterator = setup_segmentation_iterator(ome_zarr_2d, channels=_CHANNELS)

    def seg_func(img):
        result = np.zeros_like(img)
        result[..., 10:20, 10:20] = 3
        return result

    compute_segmentation(segmentation_func=seg_func, iterator=iterator)
    ome_zarr = open_ome_zarr_container(ome_zarr_2d)
    written = ome_zarr.get_label("segmentation").get_array()
    assert (written == 0).any()



from typing import Annotated

from pydantic import BaseModel, Field

from fractal_tasks_utils.transforms import (
    GaussianBlurConfig,
    HistogramEqualizationConfig,
    MedianFilterConfig,
    SizeFilterConfig,
)

PreProcess = Annotated[
    GaussianBlurConfig | MedianFilterConfig | HistogramEqualizationConfig,
    Field(discriminator="type"),
]

PostProcess = Annotated[
    SizeFilterConfig,
    Field(discriminator="type"),
]


class SegmentationTransformConfig(BaseModel):
    """Configuration for pre- and post-processing steps.

    Attributes:
        pre_process (list[PreProcess]): List of pre-processing steps.
        post_process (list[PostProcess]): List of post-processing steps.
    """

    pre_process: list[PreProcess] = Field(default_factory=list)
    post_process: list[PostProcess] = Field(default_factory=list)

    def to_pre_transforms(self) -> list:
        """Get the list of pre-processing transformations."""
        return [pre.to_transform() for pre in self.pre_process]

    def to_post_transforms(self) -> list:
        """Get the list of post-processing transformations."""
        return [post.to_transform() for post in self.post_process]



"""Tests for the measurement module."""

import pandas as pd
import pytest
from ngio import ChannelSelectionModel, create_empty_ome_zarr
from ngio.experimental.iterators import FeatureExtractorIterator

import fractal_tasks_utils.measurement
from fractal_tasks_utils.measurement import (
    compute_measurement,
    join_tables,
    setup_measurement_iterator,
)

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def ome_zarr_2d(tmp_path):
    """Minimal 2-channel 2D OME-Zarr with a derived label image."""
    zarr_path = str(tmp_path / "test.zarr")
    ome_zarr = create_empty_ome_zarr(
        store=zarr_path,
        shape=(2, 64, 64),
        pixelsize=0.5,
        channels_meta=["DAPI", "GFP"],
        axes_names=["c", "y", "x"],
    )
    ome_zarr.derive_label(name="nuclei", overwrite=True)
    return zarr_path


@pytest.fixture
def ome_zarr_with_roi_table(tmp_path):
    """Minimal 2-channel 2D OME-Zarr with a label and an image-level ROI table."""
    zarr_path = str(tmp_path / "test.zarr")
    ome_zarr = create_empty_ome_zarr(
        store=zarr_path,
        shape=(2, 64, 64),
        pixelsize=0.5,
        channels_meta=["DAPI", "GFP"],
        axes_names=["c", "y", "x"],
    )
    ome_zarr.derive_label(name="nuclei", overwrite=True)
    roi_table = ome_zarr.build_image_roi_table()
    ome_zarr.add_table(name="roi_table", table=roi_table)
    return zarr_path


def test_measurement_module_is_importable():
    assert fractal_tasks_utils.measurement is not None


def test_join_tables_single():
    table = {"label": [1, 2], "area": [10.0, 20.0], "region": ["roi_0", "roi_0"]}
    df = join_tables([table])
    assert isinstance(df, pd.DataFrame)
    assert df.index.name == "label"
    assert list(df.index) == [1, 2]
    assert list(df["area"]) == [10.0, 20.0]


def test_join_tables_multiple():
    t1 = {"label": [1, 2], "area": [10.0, 20.0], "region": ["roi_0", "roi_0"]}
    t2 = {"label": [3, 4], "area": [30.0, 40.0], "region": ["roi_1", "roi_1"]}
    df = join_tables([t1, t2])
    assert list(df.index) == [1, 2, 3, 4]
    assert list(df["area"]) == [10.0, 20.0, 30.0, 40.0]
    assert list(df["region"]) == ["roi_0", "roi_0", "roi_1", "roi_1"]


def test_join_tables_requires_at_least_one():
    with pytest.raises(ValueError):
        join_tables([])


def test_join_tables_custom_index():
    table = {"id": [10, 20], "value": [1.0, 2.0]}
    df = join_tables([table], index_key="id")
    assert df.index.name == "id"
    assert list(df.index) == [10, 20]


# ---------------------------------------------------------------------------
# setup_measurement_iterator tests
# ---------------------------------------------------------------------------


def test_setup_measurement_iterator_default(ome_zarr_2d):
    iterator = setup_measurement_iterator(ome_zarr_2d, "nuclei")
    assert isinstance(iterator, FeatureExtractorIterator)
    assert len(iterator.rois) > 0


def test_setup_measurement_iterator_channel_selection(ome_zarr_2d):
    channels = [ChannelSelectionModel(mode="label", identifier="DAPI")]
    iterator = setup_measurement_iterator(ome_zarr_2d, "nuclei", channels=channels)
    assert isinstance(iterator, FeatureExtractorIterator)
    # Iterating should yield images with only 1 channel in the last axis
    for img_chunk, _lbl_chunk, _roi in iterator.iter_as_numpy():
        assert img_chunk.shape[-1] == 1
        break


def test_setup_measurement_iterator_channels_none_includes_all(ome_zarr_2d):
    iterator = setup_measurement_iterator(ome_zarr_2d, "nuclei", channels=None)
    for img_chunk, _lbl_chunk, _roi in iterator.iter_as_numpy():
        assert img_chunk.shape[-1] == 2
        break


def test_setup_measurement_iterator_with_table(ome_zarr_with_roi_table):
    iterator = setup_measurement_iterator(
        ome_zarr_with_roi_table, "nuclei", roi_table_names=["roi_table"]
    )
    assert isinstance(iterator, FeatureExtractorIterator)
    assert len(iterator.rois) > 0


def test_setup_measurement_iterator_tables_none(ome_zarr_2d):
    iterator_no_table = setup_measurement_iterator(
        ome_zarr_2d, "nuclei", roi_table_names=None
    )
    iterator_empty = setup_measurement_iterator(
        ome_zarr_2d, "nuclei", roi_table_names=[]
    )
    assert len(iterator_no_table.rois) == len(iterator_empty.rois)


# ---------------------------------------------------------------------------
# compute_measurement tests
# ---------------------------------------------------------------------------


def test_compute_measurement_basic(ome_zarr_2d):
    iterator = setup_measurement_iterator(ome_zarr_2d, "nuclei")

    def simple_func(img, label, roi):
        return {"label": [1, 2], "area": [10.0, 20.0]}

    df = compute_measurement(measurement_func=simple_func, iterator=iterator)
    assert isinstance(df, pd.DataFrame)
    assert df.index.name == "label"
    assert "area" in df.columns


def test_compute_measurement_empty_raises(ome_zarr_2d):
    """An iterator that produces no ROIs causes compute_measurement to raise."""
    # Build an iterator and force its ROI list to be empty via product with empty ROIs
    iterator = setup_measurement_iterator(ome_zarr_2d, "nuclei")
    # Manually empty the ROI list by taking the product with an empty set
    empty_iterator = iterator.product([])

    def dummy_func(img, label, roi):
        return {"label": [], "area": []}

    with pytest.raises(ValueError):
        compute_measurement(measurement_func=dummy_func, iterator=empty_iterator)



"""Core computation logic for measurement tasks."""

import logging
from collections.abc import Callable

import numpy as np
import pandas as pd
from ngio import ChannelSelectionModel, Roi, open_ome_zarr_container
from ngio.experimental.iterators import FeatureExtractorIterator
from ngio.transforms import ZoomTransform


def join_tables(
    tables: list[dict[str, list]], index_key: str = "label"
) -> pd.DataFrame:
    """Join a list of per-ROI result dicts into a single DataFrame.

    Args:
        tables: List of dicts returned by the per-ROI extraction function.
        index_key: Column to use as the DataFrame index. Defaults to "label".

    Returns:
        A DataFrame with all ROI results concatenated and indexed by `index_key`.
    """
    if len(tables) == 0:
        raise ValueError("At least one table is required to join into a DataFrame.")
    out_dict: dict[str, list] = {}
    for table in tables:
        for key, value in table.items():
            if key not in out_dict:
                out_dict[key] = []
            out_dict[key].extend(value)
    df = pd.DataFrame(out_dict)
    df = df.set_index(index_key)
    return df


def setup_measurement_iterator(
    zarr_url: str,
    label_image_name: str,
    level_path: str | None = None,
    channels: list[ChannelSelectionModel] | None = None,
    roi_table_names: list[str] | None = None,
) -> FeatureExtractorIterator:
    """Set up a FeatureExtractorIterator for measurement tasks.

    Args:
        zarr_url: URL to the OME-Zarr container.
        label_image_name: Name of the label image to analyze.
        level_path: Optional path to a specific resolution level of the image.
            If not provided, the highest resolution level is used.
        channels: Optional list of ChannelSelectionModel to specify which channels
            to include. If None, all channels are included.
        roi_table_names: Optional list of ROI table names to include in the
            iterator. If None, no table is included.

    Returns:
        A FeatureExtractorIterator that yields (image, label, roi)
    """
    logger = logging.getLogger("fractal_tasks_utils.setup_measurement_iterator")
    logger.info(f"{zarr_url=}")

    ome_zarr = open_ome_zarr_container(zarr_url)
    logger.info(f"{ome_zarr=}")

    image = ome_zarr.get_image(path=level_path)
    logger.info(f"{image=}")

    # Get the label at the closest resolution to the image (strict=False allows
    # a ZoomTransform to handle any resolution mismatch)
    label_image = ome_zarr.get_label(
        name=label_image_name, pixel_size=image.pixel_size, strict=False
    )
    logger.info(f"{label_image=}")

    # For 2D images, squeeze the singleton z-axis (yxc); keep z for 3D (yxzc)
    axes_order = "yxc" if image.is_2d else "yxzc"

    # The ZoomTransform will handle any necessary rescaling of the label image to
    # match the image resolution (if not necessary, it will be a no-op)
    label_zoom_transform = ZoomTransform(
        input_image=label_image,
        target_image=image,
        order="nearest",
    )

    iterator = FeatureExtractorIterator(
        input_image=image,
        input_label=label_image,
        axes_order=axes_order,
        channel_selection=channels,
        label_transforms=[label_zoom_transform],
    )
    # by_zyx(strict=False): works for both 2D and 3D data
    iterator = iterator.by_zyx(strict=False)

    tables = roi_table_names if roi_table_names is not None else []
    for table_name in tables:
        table = ome_zarr.get_generic_roi_table(table_name)
        iterator = iterator.product(table)

    logger.info(f"Iterator created: {iterator=}")
    return iterator


def compute_measurement(
    *,
    measurement_func: Callable[[np.ndarray, np.ndarray, Roi], dict],
    iterator: FeatureExtractorIterator,
) -> pd.DataFrame:
    """Run the measurement computation loop.

    Iterates over ROIs using the provided iterator, applies `measurement_func`
    to each chunk, and returns all results joined into a single DataFrame.

    Args:
        measurement_func: Consumer-provided extraction function with signature
            ``(image, label, roi) -> dict``. The dict keys become DataFrame
            columns; values must be lists of equal length.
        iterator: A `FeatureExtractorIterator` (with `.by_zyx` already applied)
            that yields ``(image, label, roi)`` tuples.
    Returns:
        A DataFrame with all per-ROI results concatenated, indexed by "label".
    """
    logger = logging.getLogger("fractal_tasks_utils.compute_measurement")

    tables = []
    num_rois = len(iterator.rois)
    logging_step = max(1, num_rois // 10)

    logger.info("Starting measurement...")
    for it, (input_data, label_data, roi) in enumerate(iterator.iter_as_numpy()):
        table_dict = measurement_func(input_data, label_data, roi)
        tables.append(table_dict)

        if it % logging_step == 0 or it == num_rois - 1:
            logger.info(f"Processed ROI {it + 1}/{num_rois}")

    return join_tables(tables, index_key="label")



from fractal_tasks_utils.segmentation._compute import (
    compute_segmentation,
    setup_segmentation_iterator,
)
from fractal_tasks_utils.segmentation._models import (
    IteratorConfig,
    MaskingConfig,
)

__all__ = [
    "compute_segmentation",
    "setup_segmentation_iterator",
    "IteratorConfig",
    "MaskingConfig",
]
